#!/usr/bin/perl

use strict;
use warnings;

use open ':utf8';

use Device::Cdio::Device;
use Device::Cdio::Track;
use Digest::SHA1;
use Getopt::ArgvFile home=>1;
use Getopt::Long qw(GetOptions);
use MIME::Base64 qw(encode_base64);

my $progname = 'mbdiscid';
my $version = '0.6';

# command line options (defaulted)
my $action = 'submit';
my $browser = 'sensible-browser';
my $device = '/dev/cdrom';
my $mbserver = 'musicbrainz.org';

# command line options (undefined)
my $quiet;
my $verbose;

# disc information
my $cdio;
my $first_track_no;
my $last_track_no;
my @lbas;

# has a value if header has been output
my $header; # false

my %actions = (
  'print'  => \&action_print,
  'submit' => \&action_submit,
);

sub header {

  # do nothing if already output or user asked to be quiet
  return if (defined $header || defined $quiet);

  print STDERR "\n";
  print STDERR "$progname $version - Query MusicBrainz with computed Disc ID.\n";
  print STDERR "Copyright 2006-2009 Paul C. Bryan.\n";
  print STDERR "\n";

  $header = 1;
}

# output informative message if not quiet
sub msg {
  header; print STDERR "$_[0]\n" if (not defined $quiet);
}

# display error message if not quiet, then exit with error code
sub error {
  undef $quiet; msg "$_[0]\n"; exit 1;
}

# output diagnostic if verbosity enabled
sub diag {
  print msg $_[0] if (defined $verbose);
}

# display system call as diagnostic, then execute system call
sub sys {
  diag $_[0]; system($_[0]) == 0 or error "Error calling $_[0].";
}

# get command line options
sub get_options {

  # parse command line options (may override defaults or initialization)
  GetOptions(
    'action=s'   => \$action,
    'browser=s'  => \$browser,
    'device=s'   => \$device,
    'quiet'      => sub { $quiet = 1; undef $verbose; },
    'mbserver=s' => \$mbserver,
    'help'       => \&usage,
    'verbose'    => sub { $verbose = 1; undef $quiet; },
  ) or usage();

  error("Invalid action type: $action.") if (!defined($actions{$action}));

  # suppress header for certain actions
  $quiet = 1 if ($action eq 'print');

  # no arguments are accepted by this script
  usage() if (defined $ARGV);
}

# display usage and exit
sub usage {

  undef $quiet;
  header();

  print STDERR <<EOF;
Usage: $progname [options]

OPTIONS:
  --action=<act>    Action to perform [$action].
  --browser=<prog>  Name of program to execute as browser [$browser].
  --device=<dev>    CD-ROM device to query [$device].
  --help            Display this help message.
  --mbserver=<svr>  MusicBrainz server to query [$mbserver].
  --quiet           Quiet operation, no screen output.
  --verbose         Verbose output of progress information.

ACTIONS
  This determines what action this script should perform. The choices are:
  * print: Print the disc ID of the inserted disc.
  * submit: Submit the disc ID to MusicBrainz through browser.

CONFIGURATION FILE
  You can store persistent command-line options in a file called .$progname
  in your home directory. They will be read as defaults prior to processing
  any passed command-line options.

EOF

  exit 1;
}

sub init {
  $cdio = Device::Cdio::Device->new(-source=>$device);
}

sub get_track_lba {
  return $cdio->get_track($_[0])->get_lba();
}

# query cd table of contents
sub get_disc_info {

  msg("Getting disc information...");

  error("No tracks found in $device.") if ($cdio->get_num_tracks() == 255);

  $first_track_no = $cdio->get_first_track()->{track};
  $last_track_no = $cdio->get_last_track()->{track};

  my $data_tracks = 0;

  for my $track_no ($first_track_no .. $last_track_no) {

    my $format = $cdio->get_track($track_no)->get_format();

    if ($format eq 'audio') {
      err('audio tracks after data tracks is unsupported') if ($data_tracks != 0);
    }

    elsif ($format eq 'data') {
      $data_tracks++;
    }

    else {
      die("unrecognized track format: $format");
    }
  }

  my $leadout;

  # all audio tracks; use leadout LBA as supplied by the table of contents
  if ($data_tracks == 0) {
    $leadout = get_track_lba($last_track_no + 1);
  }

  # final data track needs to be suppressed, with calculated leadout
  elsif ($data_tracks == 1) {
    $leadout = get_track_lba($last_track_no) - 11400;
    $last_track_no--;
  }

  else {
    die("multiple data tracks unsupported");
  }

  $lbas[0] = $leadout;

  for my $track_no ($first_track_no .. $last_track_no) {
    $lbas[$track_no] = get_track_lba($track_no);
  }
}

sub get_discid {

  my $ctx = Digest::SHA1->new;
  $ctx->add(sprintf("%02X", $first_track_no));
  $ctx->add(sprintf("%02X", $last_track_no));

  for my $track_no (0 .. 99) {
    my $lba = $lbas[$track_no];
    $lba = 0 if (!defined $lba);
    $ctx->add(sprintf("%08X", $lba));
  }

  # use separate base64 encoder because sha1 base64 encoder is truncating last '=' character
  my $discid = encode_base64($ctx->digest(), '');

  # convert special http/url characters to musicbrainz replacements
  $discid =~ s/\+/./g;
  $discid =~ s/\//_/g;
  $discid =~ s/=/-/g;

  return $discid;
}

sub get_url {

  # establish first and last audio tracks on the disc
  my $tracks = $last_track_no - $first_track_no + 1;
  my $discid = get_discid();

  my $url = "http://$mbserver/bare/cdlookup.html?id=$discid&tracks=$tracks&toc=$first_track_no+$last_track_no+$lbas[0]";

  for my $track_no ($first_track_no .. $last_track_no) {
    $url = $url . "+" . $lbas[$track_no];
  }

  return $url;
}

sub action_print {
  print get_discid();
}

sub action_submit {

  # get album data from disc
  my $url = get_url();

  msg "Launching browser with Disc ID query...";

  # launch browser with url
  sys "$browser \"$url\"";

  msg "";
}

# parse command line options
get_options();

# initialize
init();

# get information about the disc
get_disc_info();

# execute the selected mode
&{$actions{$action}}();

